@@ -3,15 +3,10 @@ import traceback
 import collections
 
 import numpy as np
-from gpstime import gpsnow
+from qtpy import QtCore
+from qtpy.QtCore import Signal, Slot
 
-try:
-    from qtpy import QtCore
-    from qtpy.QtCore import Signal, Slot
-except ImportError:
-    from PyQt5 import QtCore
-    from PyQt5.QtCore import pyqtSignal as Signal
-    from PyQt5.QtCore import pyqtSlot as Slot
+from gpstime import gpsnow
 
 from . import const
 from . import nds
@@ -251,6 +246,12 @@ class DataBufferDict(object):
 class DataStore(QtCore.QObject):
     # signals
     #
+    # channel list read
+    signal_channel_list_ready = Signal()
+    # channel add/remove attempt
+    # payload: (channel, error or None)
+    signal_channel_add = Signal('PyQt_PyObject')
+    signal_channel_remove = Signal('PyQt_PyObject')
     # network data retrieval has started
     # payload: message string
     signal_data_online_start = Signal(str)
@@ -270,10 +271,9 @@ class DataStore(QtCore.QObject):
     TREND_TYPES = ['raw', 'sec', 'min']
 
     def __init__(self):
-        super(QtCore.QObject, self).__init__()
-        self.init = True
+        super().__init__()
         # use a counter to hold references to the channels, so that as
-        # many channel referneces as needed can be added, while only
+        # many channel references as needed can be added, while only
         # storing one set of channel data
         self._channels = collections.Counter()
         self.last_trend = None
@@ -282,6 +282,10 @@ class DataStore(QtCore.QObject):
         self.threads = {}
         self.reset()
         self.set_lookback(2)
+        self.available_channels = None
+        # sets up call to fetch channel list once the main loop starts
+        QtCore.QTimer.singleShot(0, self.fetch_channel_list_async)
+        logger.debug("data store initialized")
 
     def __getitem__(self, mod):
         return self.db[mod]
@@ -297,23 +301,32 @@ class DataStore(QtCore.QObject):
     def _emit_data(self, trend, online=False):
         if trend is None:
             self.signal_data.emit((None, None, online))
-        else:
+        elif self.db[trend]:
             self.signal_data.emit((self.db[trend], trend, online))
 
     ##########
 
+    def fetch_channel_list_async(self):
+        if self.available_channels:
+            return
+        self.remote_cmd('find_channels')
+
     def add_channel(self, channel):
+        if self.available_channels and channel not in self.available_channels:
+            error = f"Channel '{channel}' not found in channel list."
+            self.signal_channel_add.emit((channel, error))
+            raise ValueError(error)
+        elif const.CHANNEL_RE.match(channel) is None:
+            error = f"Invalid channel name: {channel}"
+            self.signal_channel_add.emit((channel, error))
+            raise ValueError(error)
         self._channels[channel] += 1
-        assert self._channels[channel] >= 0
+        assert self._channels[channel] > 0
+        self.signal_channel_add.emit((channel, None))
         logger.debug('CHANNEL {} {}'.format(channel, self._channels[channel]))
-        # if initializing don't fill
-        if self.init:
-            return
-        # if we already had reference to this channel emit and return
-        if self._channels[channel] > 1:
+        # if we already had reference to this channel emit
+        if self.last_trend and self._channels[channel] > 1:
             self._emit_data(self.last_trend)
-            return
-        # if online restart
         # FIXME: need to handle other online trend types
         elif self.online:
             self.online_restart()
@@ -323,13 +336,16 @@ class DataStore(QtCore.QObject):
                 if not self.db[trend]:
                     continue
                 self.remote_cmd(
-                    trend, 'extend',
-                    self.db[trend].range,
-                    channels=[channel])
+                    'extend',
+                    trend=trend,
+                    channels=[channel],
+                    start_end=self.db[trend].range,
+                )
 
     def remove_channel(self, channel):
         self._channels[channel] -= 1
         assert self._channels[channel] >= 0
+        self.signal_channel_remove.emit((channel, None))
         # FIXME: remove channel data from self.db
         logger.debug('CHANNEL {} {}'.format(channel, self._channels[channel]))
         if self.online:
@@ -337,9 +353,17 @@ class DataStore(QtCore.QObject):
 
     @property
     def channels(self):
-        # all channels with more than one referent
+        """list of all channels in the store"""
+        # the "+ Counter()" is a way to exclude elements with less
+        # than zero referrents (even though we should never hold a
+        # channel with zero referents)
         return list(self._channels + collections.Counter())
 
+    @property
+    def empty(self):
+        """True if there are no channels in the store"""
+        return len(self.channels) == 0
+
     ##########
 
     def reset(self):
@@ -356,9 +380,10 @@ class DataStore(QtCore.QObject):
         logger.debug('START ONLINE')
         self.last_trend = trend
         self.set_lookback(lookback)
+        self.stop()
         self.reset()
         self.signal_data.connect(self._online_backfill)
-        self.remote_cmd(trend, 'online', None)
+        self.remote_cmd('online', trend=trend, channels=self.channels)
 
     def _online_backfill(self, recv):
         """backfill data on online start"""
@@ -368,7 +393,7 @@ class DataStore(QtCore.QObject):
         self.signal_data.disconnect(self._online_backfill)
         start, end = self.db[trend].range
         start -= self.lookback
-        self.remote_cmd(trend, 'extendleft', (start, end))
+        self.remote_cmd('extendleft', trend=trend, channels=self.channels, start_end=(start, end))
 
     def online_stop(self):
         """stop online stream"""
@@ -377,12 +402,16 @@ class DataStore(QtCore.QObject):
             thread.stop()
 
     def online_restart(self):
-        """restart online stream"""
+        """restart online stream
+
+        Will stop the stream if there are no channels referents.
+
+        """
         if self.restart_requested:
             return
-        logger.debug('RESTART')
-        self.restart_requested = True
-        if len(list(self._channels.elements())) > 0:
+        if not self.empty:
+            logger.debug('RESTART')
+            self.restart_requested = True
             self.signal_data_online_done.connect(self._restart)
         self.online_stop()
 
@@ -406,9 +435,6 @@ class DataStore(QtCore.QObject):
         `start_end` should be a tuple of (start, end) times.
 
         """
-        if not self.channels:
-            return
-
         logger.debug("REQUEST: {} {}".format(trend, start_end))
 
         assert trend in self.TREND_TYPES
@@ -471,7 +497,7 @@ class DataStore(QtCore.QObject):
 
         # if the requested trend is empty, just get full span
         if self.db[trend] is None:
-            self.remote_cmd(trend, 'extend', (rstart, rend))
+            self.remote_cmd('extend', trend=trend, channels=self.channels, start_end=(rstart, rend))
             return
 
         # get current start/end times, adjusting inward to make sure
@@ -485,7 +511,7 @@ class DataStore(QtCore.QObject):
         if rstart >= dend or rend < dstart:
             logger.log(5, "CLEAR: {}".format(trend))
             self.db[trend] = None
-            self.remote_cmd(trend, 'extend', (rstart, rend))
+            self.remote_cmd('extend', trend=trend, channels=self.channels, start_end=(rstart, rend))
             return
 
         # emit what data we have (in case the caller is requesting
@@ -495,37 +521,50 @@ class DataStore(QtCore.QObject):
         self._emit_data(trend)
 
         if rstart < dstart:
-            self.remote_cmd(trend, 'extendleft', (rstart, dstart))
+            self.remote_cmd('extendleft', trend=trend, channels=self.channels, start_end=(rstart, dstart))
 
         if dend < rend:
-            self.remote_cmd(trend, 'extend', (dend, rend))
+            self.remote_cmd('extend', trend=trend, channels=self.channels, start_end=(dend, rend))
 
     ##########
 
-    def _command_description(self, trend, cmd=None):
+    def _command_description(self, cmd, trend):
         if cmd == 'online':
             desc = 'online '
         else:
             desc = ''
         if trend == 'sec':
-            desc += "second trend"
+            desc += "second trend data"
         elif trend == 'min':
-            desc += "minute trend"
+            desc += "minute trend data"
         elif trend == 'raw':
-            desc += "raw"
+            desc += "raw data"
         return desc
 
-    def remote_cmd(self, trend, cmd, start_end=None, channels=None):
-        # this is a thread ID used as a kind of primitive lock.  the
-        # ID should be unique enough to block requests from similar
-        # trend/action combo.  The exception is online, for which we
-        # only do one at a time.
-        if cmd == 'online':
+    def remote_cmd(self, cmd, **kwargs):
+        # the thread ID (tid) is used as a kind of primitive lock.
+        # the ID should be unique enough to block requests from
+        # similar trend/action combos.
+        if cmd == 'find_channels':
+            tid = 'find_channels'
+            desc = "channel list"
+            recv_cmd = self.remote_recv_channels
+        elif cmd == 'online':
+            if not kwargs.get('channels'):
+                return
+            trend = kwargs['trend']
             tid = 'online'
+            desc = self._command_description(cmd, trend)
+            recv_cmd = self.remote_recv_data
         else:
-            tid = '{}-{}'.format(trend, cmd)
+            if not kwargs.get('channels'):
+                return
+            trend = kwargs['trend']
+            tid = f'{cmd}-{trend}'
+            desc = self._command_description(cmd, trend)
+            recv_cmd = self.remote_recv_data
 
-        logger.debug("CMD: {} {} {} {}".format(trend, cmd, start_end, channels))
+        logger.debug(f"CMD: {cmd} {kwargs}")
         if self.active(tid):
             logger.debug("BUSY: {}".format(tid))
             return
@@ -533,30 +572,29 @@ class DataStore(QtCore.QObject):
             logger.debug("BUSY: no extend while online")
             return
 
-        msg = "Retrieving {} data...".format(self._command_description(trend, cmd))
+        msg = f"Retrieving {desc}..."
         if cmd == 'online':
             self.signal_data_online_start.emit(msg)
-        else:
-            self.signal_data_retrieve_start.emit(msg)
+        self.signal_data_retrieve_start.emit(msg)
 
-        channels = channels or self.channels
-
-        t = nds.NDSThread(tid, trend, cmd, channels, start_end)
+        t = nds.NDSThread(tid, cmd, **kwargs)
         self.threads[tid] = t
-        t.new_data.connect(self.remote_recv)
+        t.new_data.connect(recv_cmd)
         t.done.connect(self.remote_done)
         t.start()
 
     @Slot('PyQt_PyObject')
-    def remote_recv(self, recv):
+    def remote_recv_channels(self, channels):
+        nchannels = len(channels)
+        logger.info(f'channel list received: {nchannels} channels')
+        self.available_channels = channels
+        self.signal_channel_list_ready.emit()
+
+    @Slot('PyQt_PyObject')
+    def remote_recv_data(self, recv):
         logger.log(5, "")
-        logger.log(5, "RECV: {}".format(recv))
-        trend, cmd, buffers = recv
-        msg = "Receiving {} data...".format(self._command_description(trend))
-        if cmd == 'online':
-            self.signal_data_online_start.emit(msg)
-        else:
-            self.signal_data_retrieve_start.emit(msg)
+        logger.log(5, f"RECV: {recv}")
+        cmd, trend, buffers = recv
         # FIXME: should the NDS object just return this directly?
         dbd = DataBufferDict(buffers)
         dbd.set_lookback(self.lookback)
@@ -587,18 +625,15 @@ class DataStore(QtCore.QObject):
 
     @Slot('PyQt_PyObject')
     def remote_done(self, recv):
-        logger.debug("DONE: {}".format(recv))
+        logger.debug(f"DONE: {recv}")
         tid, error = recv
         if error:
-            error = "NDS error: {}".format(error)
+            error = f"NDS error ({tid}): {error}"
             logger.warning(error)
-        else:
-            error = ''
         signal = (error, self.active())
         if tid == 'online':
             self.signal_data_online_done.emit(signal)
-        else:
-            self.signal_data_retrieve_done.emit(signal)
+        self.signal_data_retrieve_done.emit(signal)
 
     def active(self, tid=None):
         """data retrieval activity status of store
@@ -630,12 +665,14 @@ class DataStore(QtCore.QObject):
             return False
         if thread.isFinished():
             return False
-        return thread.trend
+        return thread.kwargs['trend']
 
-    def terminate(self):
-        logger.debug("TERMINATE")
-        for thread in self.threads.values():
+    def stop(self):
+        logger.debug("STOP")
+        # stop data retrieval threads but not the find_channels
+        for name, thread in self.threads.items():
+            if name == 'find_channels':
+                continue
             thread.stop()
             # FIXME: thread terminate is causing problems on SL7
             # thread.terminate()
-            # self.notify_done()
